import abc, json, requests, time
from base64 import b64encode
from dataclasses import dataclass
from dataclasses_json import dataclass_json
from datetime import datetime
from enum import Enum

from .constants import * # pylint: disable=W0614
from .utilties.helper import *

ONE_CLICK_UNSUBSCRIBE = 'OneClickUnsubscribe'

class SyncDirection:
  TO_RASA = "to_rasa"
  TO_TARGET = "to_target"
  BOTH = "both"

class MessageType:
  INFO = "info"
  WARNING = "warning"
  ERROR = "error"

@dataclass_json
@dataclass
class RasaApiAttributes:
  username: str
  password: str
  key: str
  last_run_date: datetime # utc
  identity_community_guid: int
  sync_on_fieldname: str = IS_SUBSCRIBED # choose from is_active or is_subscribed
  base_url: str = "https://api.rasa.io/v1"
  sync_direction: SyncDirection = SyncDirection.BOTH

  limit: int = 1000
  keep_existing_members: bool = False
  is_person_attribute_supported: bool = False

@dataclass_json
@dataclass
class FieldMapping:
  email: str
  external_id: str = None # unique identifier field in customer database which doesn't change e.g. id
  first_name: str = None
  last_name: str = None
  is_active: str = None
  is_subscribed: str = None
  updated: str = None
  segment_code: str = None
  # exta fields can be supported by calling setattr on FieldMapping object

@dataclass_json
@dataclass
class SyncResult:
  added_to_rasa: int = 0
  updated_in_rasa: int = 0
  archived_in_rasa: int = 0
  failed_to_update_in_rasa: int = 0

class IntegrationBase(abc.ABC):
  def __init__(self, rasa_api_attrs: RasaApiAttributes, field_mapping: FieldMapping, log_func):
    """
      rasa_api_attrs: rasa api attributes
      field_mapping: field mappings
      log_func: partial function with takes 2 parameter (message, message_type)
    """
    self.rasa_api_attrs = rasa_api_attrs
    self.log_func = log_func
    self.batch_size = 25
    self.max_retry = 3
    self.attempt_delay = 2
    self.api_get_timeout = 28
    self.api_timeout = 300

    self.sync_result = SyncResult()
    self.updated_since : datetime = self.rasa_api_attrs.last_run_date
    self.field_mapping_json = field_mapping.__dict__
    self.rasa_sync_on_fieldname = self.rasa_api_attrs.sync_on_fieldname
    self.customer_sync_on_fieldname = self.field_mapping_json[self.rasa_sync_on_fieldname]

    self.rasa_api_token = self.__get_rasa_api_token()
    self.rasa_active_persons = []
    self.rasa_inactive_persons = []
    self.__rasa_persons_by_email = {}
    self.__rasa_persons_by_external_id = {}
    self.rasa_all_persons = self.__get_rasa_persons()
    if self.rasa_all_persons:
      self.log(f"Last Rasa persons found = {self.rasa_all_persons[-1]}")
    self.__current_run_rasa_persons = {}

  def log(self, message, message_type=MessageType.INFO):
    if self.log_func:
      self.log_func(message, message_type)
    else:
      print("{} | {} | {}".format(datetime.utcnow(), str(message_type), message))

  def get_rasa_api_tokens_endpoint(self):
    return "{}/tokens".format(self.rasa_api_attrs.base_url)

  def get_rasa_api_persons_endpoint(self):
    return "{}/persons".format(self.rasa_api_attrs.base_url)

  def get_rasa_api_person_endpoint(self, id):
    return  "{}/{}".format(self.get_rasa_api_persons_endpoint(), id)

  def get_auth_header(self):
    return (self.rasa_api_attrs.username, self.rasa_api_attrs.password)

  def get_basic_auth(self):
    return "Basic " + b64encode((self.rasa_api_attrs.username + ":" + self.rasa_api_attrs.password).encode()).decode()

  def __get_rasa_api_token(self):
    if self.rasa_api_attrs.key:
      body = {KEY: self.rasa_api_attrs.key}
    else:
      raise Exception("Rasa API Key is required.")
    headers = {
      "Authorization": self.get_basic_auth(),
    }
    response = self.submit_request("POST", self.get_rasa_api_tokens_endpoint(), headers=headers, payload=body)
    if (response.status_code != 201):
      raise Exception(f"Error while generating token for {body}. status_code: {response.status_code}")
    rasa_token = response.json()['results'][0][RASA_TOKEN]
    self.log("Token created. rasa-token = {}".format(rasa_token))
    return rasa_token

  def __delete_rasa_api_token(self):
    endpoint = f"{self.get_rasa_api_tokens_endpoint()}/{self.rasa_api_token}"
    headers = {
      "Authorization": self.get_basic_auth(),
    }
    response = self.submit_request("DELETE", endpoint, headers=headers)
    if (response.status_code != 204):
      self.log(f"Error while deleting token: {self.rasa_api_token}. status_code: {response.status_code}")
    else:
      self.log("Token deleted. rasa-token = {}, ".format(self.rasa_api_token))

  def sync(self) -> SyncResult:
    """
      this method syncs the data based on configuration provided to the class
    """
    try:
      self.log("sync_direction = {}".format(self.rasa_api_attrs.sync_direction))
      self.execute_pre_sync_process()

      if self.rasa_api_attrs.sync_direction == SyncDirection.TO_RASA:
        self.sync_persons_to_rasa()
        if self.get_customer_active_persons() and not self.rasa_api_attrs.keep_existing_members:
          self.__archive_rasa_persons_for_oneway_sync()
      elif self.rasa_api_attrs.sync_direction == SyncDirection.TO_TARGET:
        self.sync_persons_from_rasa()
      elif self.rasa_api_attrs.sync_direction == SyncDirection.BOTH:
        self.log("****************************************************** SYNC TO RASA ******************************************************")
        self.sync_persons_to_rasa()
        if self.get_customer_active_persons() and not self.rasa_api_attrs.keep_existing_members:
          self.__archive_rasa_persons_for_oneway_sync()
        self.log("****************************************************** SYNC FROM RASA ******************************************************")
        self.sync_persons_from_rasa()

      self.execute_post_sync_process()

      return self.sync_result
    finally:
      self.__delete_rasa_api_token()

  def sync_persons_to_rasa(self):
    self.__create_or_update_persons_in_rasa()
    self.__unsubscribe_customers_inactive_persons_in_rasa()

  def __get_rasa_person_with_attributes(self, id):
    headers = {RASA_TOKEN: self.rasa_api_token}
    endpoint = self.get_rasa_api_person_endpoint(id)
    response = self.submit_request("GET", endpoint, headers=headers)
    rasa_person = None
    if response.status_code == 200:
      results = response.json()[RESULTS]
      rasa_person = results[0].get(DATA)
    elif response.status_code == 404:
      pass # not found
    else:
      # something went wrong
      self.log("Failed to get person record with attributes. id = {}. Error = {}".format(id, self.get_rasa_response_error(response)), True)
    return rasa_person

  def __get_rasa_person_by_mapped_person(self, mapped_person):
    """
      1. search rasa person by external_id (if provided).
      2. if not found, search by email
    """
    external_id = mapped_person.get(EXTERNAL_ID)
    rasa_person = None
    if external_id:
      rasa_person = self.__rasa_persons_by_external_id.get(external_id)
    if not rasa_person and mapped_person.get(EMAIL):
      # couldn't find a person by external_id. let it search by email.
      rasa_person = self.__rasa_persons_by_email.get(mapped_person[EMAIL].lower())
    if rasa_person and self.rasa_api_attrs.is_person_attribute_supported:
      # need to get rasa_person with attribute
      rasa_person = self.__get_rasa_person_with_attributes(rasa_person.id)
    if rasa_person:
      rasa_person[UPDATED] = parse_date_from(rasa_person[UPDATED])
    return rasa_person

  def __get_rasa_persons(self):
    all_persons = []
    headers = {RASA_TOKEN: self.rasa_api_token}
    endpoint = self.get_rasa_api_persons_endpoint()
    params = {"skip" : 0, "limit": self.rasa_api_attrs.limit, "include_counts": 0}

    response = self.submit_request("GET", endpoint, headers=headers, params=params)
    for result in response.json()[RESULTS]:
      all_persons.append(result.get(DATA))
    record_count = self.get_rasa_response_metadata(response).get("record_count") or 0
    while record_count > 0:
      self.log("Rasa persons found so far = {}".format(len(all_persons)))
      params["skip"] = params["skip"] + record_count
      response = self.submit_request("GET", endpoint, headers=headers, params=params)
      for result in response.json()[RESULTS]:
        all_persons.append(result.get(DATA))
      record_count = self.get_rasa_response_metadata(response).get("record_count") or 0

    for rasa_person in all_persons:
      self.__rasa_persons_by_email[rasa_person[EMAIL].lower()] = rasa_person
      if rasa_person.get(EXTERNAL_ID):
        self.__rasa_persons_by_external_id[rasa_person[EXTERNAL_ID]] = rasa_person

      if rasa_person.get(self.rasa_sync_on_fieldname) is None or rasa_person[self.rasa_sync_on_fieldname]:
        self.rasa_active_persons.append(rasa_person)
      else:
        self.rasa_inactive_persons.append(rasa_person)

    self.log("Rasa persons found ({} = True) = {}".format(self.rasa_sync_on_fieldname, len(self.rasa_active_persons)))
    self.log("Rasa persons found ({} = False) = {}".format(self.rasa_sync_on_fieldname, len(self.rasa_inactive_persons)))
    return all_persons

  def __create_or_update_persons_in_rasa(self):
    persons_to_create_or_update = self.__get_persons_to_create_or_update()
    headers = {RASA_TOKEN: self.rasa_api_token}
    for batch in self.batchify(persons_to_create_or_update, self.batch_size):
      print(f"Processgin batch = {batch}")
      response = self.submit_request("POST", self.get_rasa_api_persons_endpoint(), headers=headers, payload=batch)
      if response.status_code not in(200, 201):
        self.log(f"Failed to create persons for batch = {batch}. Error = {self.get_rasa_response_error(response)}")
      else:
        for person_response in response.json()[RESULTS]:
          if person_response.get(ID):
            self.__current_run_rasa_persons[person_response[ID]] = True
            if person_response.get(STATUS) == CREATED:
              self.sync_result.added_to_rasa = self.sync_result.added_to_rasa + 1
            else:
              self.sync_result.updated_in_rasa = self.sync_result.updated_in_rasa + 1
          else:
            self.sync_result.failed_to_update_in_rasa = self.sync_result.failed_to_update_in_rasa + 1
            self.log(f"Couldn't find id in person_response {person_response}.")

  def __get_persons_to_create_or_update(self):
    persons_to_create_or_update = []
    mapped_persons = self.map_customer_persons_to_rasa_persons(self.get_customer_active_persons())
    for mapped_person in mapped_persons:
      rasa_person = self.__get_rasa_person_by_mapped_person(mapped_person)
      if rasa_person and not rasa_person[IS_SUBSCRIBED] and rasa_person[UNSUBSCRIBE_REASON] == ONE_CLICK_UNSUBSCRIBE and self.rasa_api_attrs.sync_direction in [SyncDirection.TO_RASA, SyncDirection.BOTH]:
        self.log("Skipping update in rasa for '{}' as this person did unsubscribed via OneClick".format(mapped_person[EMAIL]))
      elif not rasa_person:
        self.log("Creating person record in rasa for '{}'".format(mapped_person[EMAIL]))
        persons_to_create_or_update.append(self.__sanitize_mapped_rasa_person(mapped_person))
      else:
        self.__current_run_rasa_persons[rasa_person[ID]] = True
        # if customer hasn't sent updated info then assume it is new. Or if customer updated > rasa updated
        if mapped_person.get(UPDATED) is None or mapped_person[UPDATED] > rasa_person[UPDATED]:
          changes = self.__get_customer_person_changes(mapped_person, rasa_person)
          if changes:
            self.log("Updating person record in rasa for '{}'. Changes = {}".format(mapped_person[EMAIL], changes))
            persons_to_create_or_update.append(self.__sanitize_mapped_rasa_person(mapped_person))
          else:
            self.log("No change found for person record - id = {}, email = {}".format(rasa_person[ID], rasa_person[EMAIL]))
        else:
          self.log("Ignoring person record update in rasa. Rasa record is more recent - id = {}, email = {}".format(rasa_person[ID], rasa_person[EMAIL]))

    return persons_to_create_or_update


  def batchify(self, arr, batch_size):
    batches = []
    for i in range(0, len(arr), batch_size):
      batch = arr[i:i + batch_size]
      batches.append(batch)
    return batches

  def __get_customer_person_changes(self, mapped_person, rasa_person):
    result = {}
    mapped_person = self.__sanitize_mapped_rasa_person(mapped_person)
    for k, v in mapped_person.items():
      if is_different(k, v, rasa_person.get(k)):
        result[k] = v
    return result

  def __sanitize_mapped_rasa_person(self, mapped_person):
    mapped_person.pop("updated", None)
    return mapped_person

  def __unsubscribe_customers_inactive_persons_in_rasa(self):
    mapped_persons = self.map_customer_persons_to_rasa_persons(self.get_customer_inactive_persons())
    if mapped_persons:
      for mapped_person in mapped_persons:
        rasa_person = self.__get_rasa_person_by_mapped_person(mapped_person)
        if rasa_person and rasa_person[self.rasa_sync_on_fieldname]:
          if mapped_person.get(UPDATED) is None or mapped_person[UPDATED] > rasa_person[UPDATED]:
            self.__unsubscribe_rasa_person(rasa_person)

  def __unsubscribe_rasa_person(self, rasa_person):
    self.log("Unsubscribing person record in rasa with id = '{}', email = '{}'".format(rasa_person[ID], rasa_person[EMAIL]))
    headers = {RASA_TOKEN: self.rasa_api_token}
    endpoint = self.get_rasa_api_person_endpoint(rasa_person[ID])
    response = self.submit_request("PUT", endpoint, headers=headers, payload={self.rasa_sync_on_fieldname:0})
    if response.status_code != 200:
      self.log("Failed to unsubscribe rasa person record. id = {}. Error = {}".format(rasa_person[ID], self.get_rasa_response_error(response)))

  def __archive_rasa_person(self, rasa_person):
    self.log(f"Archiving person record in rasa with id = '{rasa_person[ID]}', email = '{rasa_person[EMAIL]}'")
    headers = {RASA_TOKEN: self.rasa_api_token}
    endpoint = self.get_rasa_api_person_endpoint(rasa_person[ID])
    payload = {IS_ARCHIVED:1}
    if rasa_person.get(EXTERNAL_ID):
      payload[EXTERNAL_ID] = rasa_person[EXTERNAL_ID]
    response = self.submit_request("PUT", endpoint, headers=headers, payload=payload)
    if response.status_code != 200:
      self.log("Failed to archive rasa person record. id = {}. Error = {}".format(rasa_person[ID], self.get_rasa_response_error(response)))
    else:
      self.sync_result.archived_in_rasa = self.sync_result.archived_in_rasa + 1

  def __archive_rasa_persons_for_oneway_sync(self):
    print(f"rasa_all_persons count = {len(self.rasa_all_persons)}")
    print(f"current_run_rasa_persons count = {len(self.__current_run_rasa_persons)}")
    for rasa_person in self.rasa_all_persons:
      if not self.__current_run_rasa_persons.get(rasa_person[ID]):
        self.__archive_rasa_person(rasa_person)

  def map_customer_persons_to_rasa_persons(self, customer_persons):
    mapped_persons = []
    for customer_person in customer_persons:
      mapped_person = {k : customer_person[v] for k, v in self.field_mapping_json.items() if v and customer_person.get(v) is not None}
      if mapped_person.get(UPDATED):
        mapped_person[UPDATED] = parse_date_from(mapped_person[UPDATED])
      mapped_persons.append(mapped_person)
    return mapped_persons

  def map_rasa_persons_to_customer_persons(self, rasa_persons):
    mapped_persons = []
    for rasa_person in rasa_persons:
      if self.rasa_api_attrs.is_person_attribute_supported:
        rasa_person = self.__get_rasa_person_with_attributes(rasa_person[ID])
        if not rasa_person:
          continue

      mapped_person = {v : rasa_person[k] for k, v in self.field_mapping_json.items() if v and rasa_person.get(k) is not None}
      if self.field_mapping_json.get(UPDATED) and mapped_person.get(self.field_mapping_json[UPDATED]):
        mapped_person[self.field_mapping_json[UPDATED]] = parse_date_from(mapped_person[self.field_mapping_json[UPDATED]])
      mapped_persons.append(mapped_person)

    return mapped_persons

  def __get_rasa_id_from_create_response(self, response):
    results = response.json()[RESULTS]
    return results[0].get(ID)

  def get_rasa_response_metadata(self, response):
    try:
      return response.json()["metadata"]
    except:
      return {}

  def get_rasa_response_error(self, response):
    try:
      return response.json()['metadata']["errors"]
    except:
      return "Error response not found"

  def submit_request(self, method, endpoint, headers = {}, payload = {}, params={}, attempt = 1):
    self.log(f"{method}-ing to {endpoint}: payload={payload}, params={params}")
    timeout = self.api_get_timeout if method.upper() == "GET" else self.api_timeout

    r = requests.request(method, endpoint, headers=headers, json=payload, params=params, timeout=timeout)
    if not r.ok:
      if attempt <= self.max_retry:
        retry_delay = attempt * self.attempt_delay
        self.log(f"Resubmitting attempt #{attempt} after delay of {retry_delay}")
        time.sleep(retry_delay)
        return self.submit_request(method, endpoint, headers=headers, payload=payload, params=params, attempt=attempt+1)
      else:
        self.log(f"Request Timeout after attempt #{attempt}. Response = {r.status_code} {r.text}")

    return r


  @abc.abstractmethod
  def get_customer_active_persons(self):
    pass # should retuns array

  @abc.abstractmethod
  def get_customer_inactive_persons(self):
    pass # should retuns array

  @abc.abstractmethod
  def sync_persons_from_rasa(self):
    pass

  @abc.abstractmethod
  def execute_pre_sync_process(self):
    pass

  @abc.abstractmethod
  def execute_post_sync_process(self):
    pass


import json, pytz
from datetime import datetime

BOOLEAN_TRUE_VALUES = [True, 1, "true", "t", "yes", "y", "1"]
BOOLEAN_FALSE_VALUES = [False, 0, "false", "f", "no", "n", "0"]
BOOLEAN_VALUES = BOOLEAN_TRUE_VALUES + BOOLEAN_FALSE_VALUES

def is_truthy(value):
  return not is_falsey(value)

def is_falsey(value):
  """ Falsey values - None, False, "False", "No", 0, "" """
  if value == None or value == False or value == 0:
    return True
  value_string = str(value).lower()
  return len(value_string) == 0 or value_string in BOOLEAN_FALSE_VALUES

def parse_date_from(input_date, format=None):
  if input_date:
    if isinstance(input_date, datetime):
      return input_date
    else:
      for fmt in [format, "%m/%d/%Y %I:%M:%S %p", "%Y-%m-%d %H:%M:%S.%f", "%Y-%m-%d %H:%M:%S", "%Y-%m-%d-%H-%M", "%Y-%m-%d",
        "%Y-%m-%dT%H:%M:%S.%f", "%Y-%m-%dT%H:%M:%S.%f%z"]:
        try:
          parsed_date = datetime.strptime(input_date, fmt)
          return parsed_date if parsed_date.tzinfo else parsed_date.replace(tzinfo=pytz.UTC)
        except:
          pass
  return None

def is_different(key, from_value, to_value):
  if from_value == None and to_value == None:
    return False
  if (from_value != None and to_value == None) or (to_value != None and from_value == None):
    if key == "segment_code":
      if from_value or to_value:
        return True # one of them is not empty
      else:
        return False
    else:
      return True # if only one of them is None, then they are different
  if is_boolean(from_value):
    return convert_to_boolean(from_value) != convert_to_boolean(to_value)
  if isinstance(from_value, str) and isinstance(to_value, str):
    # for strings type, let it be CASE INSENSITIVE
    return from_value.lower().strip() != to_value.lower().strip()
  # TODO date support. leaving for now

  return from_value != to_value

def is_boolean(value):
  if value in (BOOLEAN_VALUES) or str(value).lower() in (BOOLEAN_VALUES):
    return True
  else:
    return False

def convert_to_boolean(value):
  if value in BOOLEAN_TRUE_VALUES or str(value).lower() in BOOLEAN_TRUE_VALUES:
    return True
  elif value in BOOLEAN_FALSE_VALUES or str(value).lower() in BOOLEAN_FALSE_VALUES:
    return False
  else:
    return None



import json, pytz
from dataclasses import dataclass
from dataclasses_json import dataclass_json
from datetime import datetime
from simple_salesforce import Salesforce

SF_DATETIME_FORMAT = '%Y-%m-%dT%H:%M:%S.%f%z'
SF_CONTACTS_SUPPORTED_FIELDS = " Id, AccountId, FirstName, LastName, Email, Title, Name, Description, Phone, PhotoUrl, LastModifiedDate "


@dataclass_json
@dataclass
class SFApiAttributes:
  username: str
  password: str
  security_token: str
  sync_filter_field: str = None # a boolean field to filter a set of record to participate in sync

class RasaSalesforce:
  def __init__(self, sf_api_attrs:SFApiAttributes, customer_sync_on_fieldname):
    self.__instance = Salesforce(username=sf_api_attrs.username, password=sf_api_attrs.password, security_token=sf_api_attrs.security_token)
    self.customer_sync_on_fieldname = customer_sync_on_fieldname
    self.sync_filter_field = sf_api_attrs.sync_filter_field

  def get_person_by_id(self, id):
    try:
      query = "SELECT {}, {} FROM Contact WHERE id='{}' ORDER BY CreatedDate ASC LIMIT 1".format(
        SF_CONTACTS_SUPPORTED_FIELDS,
        self.customer_sync_on_fieldname,
        id
      )
      return self.__get_one_record(query)
    except Exception:
      return None


  def get_person_by_email(self, email):
    query = "SELECT {}, {} FROM Contact WHERE Email='{}' ORDER BY CreatedDate ASC LIMIT 1".format(
        SF_CONTACTS_SUPPORTED_FIELDS,
        self.customer_sync_on_fieldname,
        email
      )
    return self.__get_one_record(query)

  def __get_persons(self, active, last_modified=None):
    last_modified = self.convert_date_to_salesforce_format(last_modified)
    query = "SELECT {}, {} FROM Contact WHERE Email!='' AND {}={} AND LastModifiedDate > {} ".format(
        SF_CONTACTS_SUPPORTED_FIELDS,
        self.customer_sync_on_fieldname,
        self.customer_sync_on_fieldname,
        active,
        last_modified
      )
    if self.sync_filter_field:
      query += " AND {}=true ".format(self.sync_filter_field)

    query += " ORDER BY CreatedDate ASC"
    r =  self.__get_all_records(query)
    return r

  def get_active_persons(self, last_modified=None):
    return self.__get_persons("true", last_modified)

  def get_inactive_persons(self, last_modified=None):
    return self.__get_persons("false", last_modified)

  def create_person(self, data):
    return self.__instance.Contact.create(data)

  def update_person(self, record_id, data):
    return self.__instance.Contact.update(record_id, data)

  def delete_person(self, record_id):
    return self.__instance.Contact.delete(record_id)

  def convert_date_to_salesforce_format(self, input_date):
    if not input_date:
      input_date = datetime.strptime("1900-01-01T00:00:00.000+0000", SF_DATETIME_FORMAT)
    if input_date.tzinfo is not None and input_date.tzinfo.utcoffset(input_date) is not None:
      return input_date.isoformat() # this is timezone aware
    else:
      return input_date.replace(tzinfo=pytz.UTC).isoformat() # naive datetime

  def __get_one_record(self, query, include_deleted=False):
    response = self.__instance.query(query, include_deleted=include_deleted)
    if response.get("done"):
      records = response.get("records")
      if records:
        record = records[0]
        return {k : record[k] for k in record if k != 'attributes'}
      else:
        return None
    else:
      raise Exception("Something went wrong while fetching one record from salesforce. Query = {}".format(query))

  def __get_all_records(self, query, include_deleted=False):
    batch_size = 50
    page_number = 0
    all_records = []
    while True:
      response = self.__instance.query("{} limit {} offset {}".format(query, batch_size, page_number * batch_size),
        include_deleted=include_deleted)
      if response.get("done"):
        records = response.get("records")
        for record in records:
          all_records.append({k : record[k] for k in record if k != 'attributes'})
        page_number += 1
        if len(records) < batch_size:
          return all_records
      else:
        raise Exception("Something went wrong while fetching records from salesforce. Query = {}".format(query))
    return all_records



import os, sys, traceback, unittest
from datetime import datetime, timedelta
from rasaio.integrations import * # pylint: disable=W0614
from rasaio.integrations.salesforce import SalesforceIntegration, SFApiAttributes

class TestSalesforce(unittest.TestCase):
  rasa_api_attrs = RasaApiAttributes(
    username = os.environ.get("rasa_username"),
    password = os.environ.get("rasa_password"),
    key = os.environ.get("rasa_key"),
    base_url = "https://api-dev.rasa.io/v1",
    last_run_date =  (datetime.utcnow() - timedelta(days=1))
  )

  field_mapping = FieldMapping(
    email = "Email",
    external_id = "Id",
    first_name = "FirstName",
    last_name = "LastName",
    updated = "LastModifiedDate",
    is_subscribed = "RasaSubscribed__c"
  )

  sf_api_attrs = SFApiAttributes(
    username=os.environ.get("salesfoce_username"),
    password=os.environ.get("salesfoce_password"),
    security_token=os.environ.get("salesfoce_security_token")
  )

  def test_get_persons(self):
    try:
      sf = SalesforceIntegration(self.rasa_api_attrs, self.sf_api_attrs, self.field_mapping)
      sf.sync()
      self.assertTrue(True)
    except Exception:
      traceback.print_exc()
      self.assertTrue(False)


if __name__ == '__main__':
  unittest.main()


RASA_DATETIME_FORMAT = "%Y-%m-%d %H:%M:%S"

DATA = "data"
RESULTS = "results"

# API_ENDPOINT = "api_endpoint"
KEY = "key"
RASA_TOKEN = "rasa-token"
COMMUNITY_ID = "community_id"
TYPE = "type"

# # Person fields
ID = "id"
EXTERNAL_ID = "external_id"
EMAIL = "email"
FIRST_NAME = "first_name"
LAST_NAME = "last_name"
UPDATED = "updated"
IS_SUBSCRIBED = "is_subscribed"
UNSUBSCRIBE_REASON = "unsubscribe_reason"
IS_ACTIVE = "is_active"
IS_ARCHIVED = "is_archived"
STATUS = "status"
CREATED = "created"



import os
from setuptools import setup, find_packages

try: # for pip >= 10
  from pip._internal.req import parse_requirements
except ImportError: # for pip <= 9.0.3
  from pip.req import parse_requirements

install_reqs = parse_requirements('requires.txt', session=False)

try:
  requirements = [str(ir.req) for ir in install_reqs]
except:
  requirements = [str(ir.requirement) for ir in install_reqs]

README="""Python class to integrate rasa.io and another customer database"""

setup(
  name='rasaio.integrations',
  version='1.0.12',
  description=README,
  long_description=README,
  long_description_content_type='text/x-rst',
  classifiers=[
    "Programming Language :: Python :: 3.6"
  ],
  author="rasa.io",
  author_email="chandra.bhushan@rasa.io",
  packages=find_packages(),
  python_requires='>=3.6, <4',
  url="https://bitbucket.org/rasa-io/rasa-integrations.git",
  install_requires=requirements,
  include_package_data=True,
  zip_safe=True
)



import json, traceback
from dataclasses import dataclass
from dataclasses_json import dataclass_json
from datetime import datetime
from .. import *
from ..utilties.helper import *
from .salesforce import RasaSalesforce, SFApiAttributes

SALESFORCE_ID = "Id"
# TODO use mapped field
SALESFORCE_FIRSTNAME = "FirstName"
SALESFORCE_LASTNAME = "LastName"
SALESFORCE_EMAIL = "Email"
SALESFORCE_DATE_UPDATED = "LastModifiedDate"

class SalesforceIntegration(IntegrationBase):
  def __init__(self, rasa_api_attrs: RasaApiAttributes, sf_api_attrs: SFApiAttributes, field_mapping: FieldMapping, log_func= None):
    super(__class__, self).__init__(rasa_api_attrs, field_mapping, log_func)
    self.sf_api_attrs : SFApiAttributes = sf_api_attrs
    self.rasa_salesforce = RasaSalesforce(sf_api_attrs, self.customer_sync_on_fieldname)
    self.__populate_initial_data()

  def __populate_initial_data(self):
    self.__active_persons = self.rasa_salesforce.get_active_persons(self.updated_since)
    self.log("Customer persons found (subscribed) = {}".format(len(self.__active_persons)))
    self.__inactive_persons = self.rasa_salesforce.get_inactive_persons(self.updated_since)
    self.log("Customer persons found (unsubscribed) = {}".format(len(self.__inactive_persons)))

  def get_customer_active_persons(self):
    return self.__active_persons

  def get_customer_inactive_persons(self):
    # overridden method
    return self.__inactive_persons

  def sync_persons_from_rasa(self):
    self.__create_or_update_rasa_active_persons()
    self.__unsubscribe_rasa_inactive_persons()

  def __create_or_update_rasa_active_persons(self):
    mapped_persons = self.map_rasa_persons_to_customer_persons(self.rasa_active_persons)
    for mapped_person in mapped_persons:
      mapped_person[self.customer_sync_on_fieldname] = "true"
      sf_person = None
      if mapped_person.get(SALESFORCE_ID):
        sf_person = self.rasa_salesforce.get_person_by_id(mapped_person[SALESFORCE_ID])
      else:
        sf_person = self.rasa_salesforce.get_person_by_email(mapped_person[SALESFORCE_EMAIL])
      try:
        if not sf_person:
          if mapped_person.get(SALESFORCE_FIRSTNAME) and mapped_person.get(SALESFORCE_LASTNAME):
            self.log("Creating person record in salesforcce for '{}'".format(mapped_person[SALESFORCE_EMAIL]))
            self.rasa_salesforce.create_person(self.__sanitize_mapped_person(mapped_person))
          else:
            self.log("Skipping person create in Salesforce for '{}'. FirstName/LastName missing.".format(mapped_person[SALESFORCE_EMAIL]))
        else:
          if mapped_person[SALESFORCE_DATE_UPDATED] > parse_date_from(sf_person[SALESFORCE_DATE_UPDATED]):
            changes = self.__get_rasa_person_changes(mapped_person, sf_person)
            if changes:
              self.log("Updating person record in Salesforce for '{}'. Changes = {}".format(mapped_person[SALESFORCE_EMAIL], changes))
              self.rasa_salesforce.update_person(sf_person[SALESFORCE_ID], changes)
            else:
              self.log("No change found for person record - id = {}, email = {}".format(sf_person[SALESFORCE_ID], sf_person[SALESFORCE_EMAIL]))
          else:
            self.log("Ignoring person record update in Salesforce. Salesforce record is more recent - id = {}, email = {}".format(sf_person[SALESFORCE_ID], sf_person[SALESFORCE_EMAIL]))
      except Exception as ex:
        # log this and continue for next record
        traceback.print_exc()
        self.log("Failed to add/update person in Salesforce for {}. Error = {}".format(mapped_person[SALESFORCE_EMAIL], ex), MessageType.ERROR)

  def __get_rasa_person_changes(self, mapped_person, sf_person):
    result = {}
    mapped_person = self.__sanitize_mapped_person(mapped_person)
    for k, v in mapped_person.items():
      if is_different(k, v, sf_person.get(k)):
        result[k] = v
    return result

  def __unsubscribe_rasa_inactive_persons(self):
    mapped_persons = self.map_rasa_persons_to_customer_persons(self.rasa_inactive_persons)
    for mapped_person in mapped_persons:
      sf_person = None
      if mapped_person.get(SALESFORCE_ID):
        sf_person = self.rasa_salesforce.get_person_by_id(mapped_person[SALESFORCE_ID])
      else:
        sf_person = self.rasa_salesforce.get_person_by_email(mapped_person[SALESFORCE_EMAIL])
      if sf_person and not sf_person[self.customer_sync_on_fieldname] \
          and (not mapped_person[SALESFORCE_DATE_UPDATED] or mapped_person[SALESFORCE_DATE_UPDATED] > parse_date_from(sf_person[SALESFORCE_DATE_UPDATED])):
        self.log("Unsubscribing person record in Salesforce for '{}'".format(sf_person[SALESFORCE_EMAIL]))
        self.rasa_salesforce.update_person(sf_person[SALESFORCE_ID], {self.customer_sync_on_fieldname: False})
      else:
        self.log("Either record doesn't exist in Salesforce or already unsubscribed or more recent. Email = {}".format(mapped_person[SALESFORCE_EMAIL]))

  def __sanitize_mapped_person(self, mapped_person):
    mapped_person.pop("LastModifiedDate", None)
    return mapped_person

  def execute_pre_sync_process(self):
    pass

  def execute_post_sync_process(self):
    pass
